from .catalogue import post_processor_catalogue, register_post_processor, remove_post_processor

from .base import PostProcessor
from .filth_replacer import FilthReplacer
from .prefix_suffix import PrefixSuffixReplacer
from .remover import FilthRemover
